@@ -27,8 +27,11 @@ import java.io.IOException;
 import java.io.InputStream;
 import java.io.OutputStream;
 import java.io.PrintStream;
+import java.net.InetAddress;
+import java.net.NetworkInterface;
 import java.net.SocketTimeoutException;
 import java.util.HashSet;
+import java.util.Enumeration;
 import java.util.Iterator;
 
 import org.apache.commons.net.io.FromNetASCIIOutputStream;
@@ -75,8 +78,6 @@ import org.apache.commons.net.io.ToNetASCIIInputStream;
  *
  * </code>
  *
- *
- * @author <A HREF="mailto:daniel.armbrust.list@gmail.com">Dan Armbrust</A>
  * @since 2.0
  */
 
@@ -91,6 +92,7 @@ public class TFTPServer implements Runnable
     private File serverReadDirectory_;
     private File serverWriteDirectory_;
     private final int port_;
+    private final InetAddress laddr_;
     private Exception serverException = null;
     private final ServerMode mode_;
 
@@ -150,6 +152,7 @@ public class TFTPServer implements Runnable
      *
      * @param serverReadDirectory directory for GET requests
      * @param serverWriteDirectory directory for PUT requests
+     * @param port the port to use
      * @param mode A value as specified above.
      * @param log Stream to write log message to. If not provided, uses System.out
      * @param errorLog Stream to write error messages to. If not provided, uses System.err.
@@ -162,13 +165,88 @@ public class TFTPServer implements Runnable
         mode_ = mode;
         log_ = (log == null ? nullStream: log);
         logError_ = (errorLog == null ? nullStream : errorLog);
+        laddr_ = null;
+        launch(serverReadDirectory, serverWriteDirectory);
+    }
+
+    /**
+     * Start a TFTP Server on the specified port. Gets and Puts occur in the specified directory.
+     *
+     * The server will start in another thread, allowing this constructor to return immediately.
+     *
+     * If a get or a put comes in with a relative path that tries to get outside of the
+     * serverDirectory, then the get or put will be denied.
+     *
+     * GET_ONLY mode only allows gets, PUT_ONLY mode only allows puts, and GET_AND_PUT allows both.
+     * Modes are defined as int constants in this class.
+     *
+     * @param serverReadDirectory directory for GET requests
+     * @param serverWriteDirectory directory for PUT requests
+     * @param port The local port to bind to.
+     * @param localaddr The local address to bind to.
+     * @param mode A value as specified above.
+     * @param log Stream to write log message to. If not provided, uses System.out
+     * @param errorLog Stream to write error messages to. If not provided, uses System.err.
+     * @throws IOException if the server directory is invalid or does not exist.
+     */
+    public TFTPServer(File serverReadDirectory, File serverWriteDirectory, int port,
+        InetAddress localaddr, ServerMode mode, PrintStream log, PrintStream errorLog)
+        throws IOException
+    {
+        port_ = port;
+        mode_ = mode;
+        laddr_ = localaddr;
+        log_ = (log == null ? nullStream: log);
+        logError_ = (errorLog == null ? nullStream : errorLog);
+        launch(serverReadDirectory, serverWriteDirectory);
+    }
+
+    /**
+     * Start a TFTP Server on the specified port. Gets and Puts occur in the specified directory.
+     *
+     * The server will start in another thread, allowing this constructor to return immediately.
+     *
+     * If a get or a put comes in with a relative path that tries to get outside of the
+     * serverDirectory, then the get or put will be denied.
+     *
+     * GET_ONLY mode only allows gets, PUT_ONLY mode only allows puts, and GET_AND_PUT allows both.
+     * Modes are defined as int constants in this class.
+     *
+     * @param serverReadDirectory directory for GET requests
+     * @param serverWriteDirectory directory for PUT requests
+     * @param port the port to use
+     * @param localiface The local network interface to bind to.
+     *  The interface's first address wil be used.
+     * @param mode A value as specified above.
+     * @param log Stream to write log message to. If not provided, uses System.out
+     * @param errorLog Stream to write error messages to. If not provided, uses System.err.
+     * @throws IOException if the server directory is invalid or does not exist.
+     */
+    public TFTPServer(File serverReadDirectory, File serverWriteDirectory, int port,
+        NetworkInterface localiface, ServerMode mode, PrintStream log, PrintStream errorLog)
+        throws IOException
+    {
+        mode_ = mode;
+        port_= port;
+        InetAddress iaddr = null;
+        if (localiface != null)
+        {
+            Enumeration<InetAddress> ifaddrs = localiface.getInetAddresses();
+            if (ifaddrs != null)
+            {
+                if (ifaddrs.hasMoreElements()) iaddr = ifaddrs.nextElement();
+            }
+        }
+        log_ = (log == null ? nullStream: log);
+        logError_ = (errorLog == null ? nullStream : errorLog);
+        laddr_ = iaddr;
         launch(serverReadDirectory, serverWriteDirectory);
     }
 
     /**
      * Set the max number of retries in response to a timeout. Default 3. Min 0.
      *
-     * @param retries
+     * @param retries number of retries, must be &gt; 0
      */
     public void setMaxTimeoutRetries(int retries)
     {
@@ -181,6 +259,7 @@ public class TFTPServer implements Runnable
 
     /**
      * Get the current value for maxTimeoutRetries
+     * @return the max allowed number of retries
      */
     public int getMaxTimeoutRetries()
     {
@@ -189,8 +268,9 @@ public class TFTPServer implements Runnable
 
     /**
      * Set the socket timeout in milliseconds used in transfers. Defaults to the value here:
-     * http://commons.apache.org/net/apidocs/org/apache/commons/net/tftp/TFTP.html#DEFAULT_TIMEOUT
+     * https://commons.apache.org/net/apidocs/org/apache/commons/net/tftp/TFTP.html#DEFAULT_TIMEOUT
      * (5000 at the time I write this) Min value of 10.
+     * @param timeout the timeout; must be larger than 10
      */
     public void setSocketTimeout(int timeout)
     {
@@ -203,6 +283,7 @@ public class TFTPServer implements Runnable
 
     /**
      * The current socket timeout used during transfers in milliseconds.
+     * @return the timeout value
      */
     public int getSocketTimeout()
     {
@@ -240,7 +321,11 @@ public class TFTPServer implements Runnable
         // we want the server thread to listen forever.
         serverTftp_.setDefaultTimeout(0);
 
-        serverTftp_.open(port_);
+        if (laddr_ != null) {
+            serverTftp_.open(port_, laddr_);
+        } else {
+            serverTftp_.open(port_);
+        }
 
         serverThread = new Thread(this);
         serverThread.setDaemon(true);
@@ -269,7 +354,7 @@ public class TFTPServer implements Runnable
         return !shutdownServer;
     }
 
-//    @Override
+    @Override
     public void run()
     {
         try
@@ -371,12 +456,12 @@ public class TFTPServer implements Runnable
             }
         }
 
-//        @Override
+        @Override
         public void run()
         {
             try
             {
-                transferTftp_ = new TFTP();
+                transferTftp_ = newTFTP();
 
                 transferTftp_.beginBufferedOps();
                 transferTftp_.setDefaultTimeout(socketTimeout_);
@@ -490,7 +575,7 @@ public class TFTPServer implements Runnable
 
                         lastSentData = new TFTPDataPacket(trrp.getAddress(), trrp.getPort(), block,
                                 temp, 0, readLength);
-                        transferTftp_.bufferedSend(lastSentData);
+                        sendData(transferTftp_, lastSentData); // send the data
                     }
 
                     answer = null;
@@ -631,7 +716,7 @@ public class TFTPServer implements Runnable
                 }
 
                 TFTPAckPacket lastSentAck = new TFTPAckPacket(twrp.getAddress(), twrp.getPort(), 0);
-                transferTftp_.bufferedSend(lastSentAck);
+                sendData(transferTftp_, lastSentAck); // send the data
 
                 while (true)
                 {
@@ -706,7 +791,7 @@ public class TFTPServer implements Runnable
                         }
 
                         lastSentAck = new TFTPAckPacket(twrp.getAddress(), twrp.getPort(), block);
-                        transferTftp_.bufferedSend(lastSentAck);
+                        sendData(transferTftp_, lastSentAck); // send the data
                         if (dataLength < TFTPDataPacket.MAX_DATA_LENGTH)
                         {
                             // end of stream signal - The tranfer is complete.
@@ -845,7 +930,7 @@ public class TFTPServer implements Runnable
     /**
      * Set the stream object to log debug / informational messages. By default, this is a no-op
      *
-     * @param log
+     * @param log the stream to use for logging
      */
     public void setLog(PrintStream log)
     {
@@ -855,10 +940,24 @@ public class TFTPServer implements Runnable
     /**
      * Set the stream object to log error messsages. By default, this is a no-op
      *
-     * @param logError
+     * @param logError the stream to use for logging errors
      */
     public void setLogError(PrintStream logError)
     {
         this.logError_ = logError;
     }
+
+    /*
+     * Allow test code to customise the TFTP instance
+     */
+    TFTP newTFTP() {
+        return new TFTP();
+    }
+
+    /*
+     * Also allow customisation of sending data/ack so can generate errors if needed
+     */
+    void sendData(TFTP tftp, TFTPPacket data) throws IOException {
+        tftp.bufferedSend(data);
+    }
 }
